// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#ifndef _SINGLEENDMERGER_H_
#define _SINGLEENDMERGER_H_

#include "hobbesconf.h"
#include "search/read.h"
#include "search/merger.h"
#include "search/singleendverifier.h"

#include <vector>
#include <string>

class Indexer;

class SingleEndMerger: public Merger {
public:
	std::vector<unsigned> tmp;
	std::vector<unsigned> tmp1;

	SingleEndMerger(Indexer* indexer, unsigned distThresh) :
			Merger(indexer, distThresh) {
		tmp.reserve(TMPVEC_PREALLOC_SIZE);
		tmp1.reserve(TMPVEC_PREALLOC_SIZE);
	}

	// generating candidates using k + 1 prefix
	// filtering candidates using bit vectors
	// implementations of derived classes are in singleendmerger.cc
	virtual void findCandidatesAll(Read& read, std::vector<unsigned>& candidates) = 0;

	// generating candidates using k + 2 prefix with BitVector inverted lists
	// k + 1 prefix generates candidates and 1 additional inverted list filters candidates.
	// (this function does not use bit vectors at all)
	// implementations of derived classes are in singleendmergerplus.cc
	virtual void findCandidatesAll( const std::vector<BitVectorInvertedListWithOffset>& invLists,
					const std::vector<unsigned>& prefixGrams,
					std::vector<unsigned>& candidates) = 0;

	// generating candidates using k + 2 prefix with normal inverted lists
	// k + 1 prefix generates candidates and 1 additional inverted list filters candidates.
	// implementations of derived classes are in singleendmergerplus.cc
	virtual void findCandidatesAll( const std::vector<InvertedListWithOffset>& invLists,
					const std::vector<unsigned>& prefixGrams,
					std::vector<unsigned>& candidates) = 0;
	// generating candidates by taking union of inverted lists of prefix grams.
	// NF stands for no filtering :(
	// implementations of derived classes are in singleendmergerplus.cc
	virtual void findCandidatesAllNF(const std::vector<InvertedListWithOffset>& invLists,
					 const std::vector<unsigned>& prefixGrams,
					 std::vector<unsigned>& candidates) = 0;

	virtual bool findCandidatesAndMappingsK(unsigned& k, Read& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results,
			SingleEndVerifier& verifier) = 0;

	virtual bool findCandidatesAndExactMappingsK(unsigned& k, Read& read,
			std::vector<unsigned>& candidates, std::vector<ReadMapping>& results,
			SingleEndVerifier& verifier) = 0;
};

class SingleEndIndelMerger: public SingleEndMerger {
public:
	SingleEndIndelMerger(Indexer* indexer, unsigned distThresh) :
			SingleEndMerger(indexer, distThresh) {
	}

	void findCandidatesAll(Read& read, std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<BitVectorInvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<InvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates); 

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAllNF(const std::vector<InvertedListWithOffset>& invLists,
				 const std::vector<unsigned>& prefixGrams,
				 std::vector<unsigned>& candidates);

	bool findCandidatesAndMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);

	bool findCandidatesAndExactMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);
};

class SingleEndHeuristicIndelMerger: public SingleEndMerger {
public:
	SingleEndHeuristicIndelMerger(Indexer* indexer, unsigned distThresh) :
			SingleEndMerger(indexer, distThresh) {
		if (distThresh >= 4) {
			bitVectorThresh = 4;
		} else {
			bitVectorThresh = distThresh;
		}

		delta = distThresh; //distThresh/2 + distThresh%2;
	}

	void findCandidatesAll(Read& read, std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<BitVectorInvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<InvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates); 

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAllNF(const std::vector<InvertedListWithOffset>& invLists,
				 const std::vector<unsigned>& prefixGrams,
				 std::vector<unsigned>& candidates);

	bool findCandidatesAndMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);

	bool findCandidatesAndExactMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);

private:
	int delta;
	bool rgt(int x, int y){ return (x - y) > delta; }
	bool rlt(int x, int y){ return (y - x) > delta; }

private:
	unsigned bitVectorThresh;
};

class SingleEndHammingMerger: public SingleEndMerger {
public:
	SingleEndHammingMerger(Indexer* indexer, unsigned distThresh) :
			SingleEndMerger(indexer, distThresh) {
	}

	void findCandidatesAll(Read& read, std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<BitVectorInvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates);

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAll( const std::vector<InvertedListWithOffset>& invLists,
				const std::vector<unsigned>& prefixGrams,
				std::vector<unsigned>& candidates); 

	// implementation of this function is inculded in singleendmergerplus.cc
	void findCandidatesAllNF(const std::vector<InvertedListWithOffset>& invLists,
				 const std::vector<unsigned>& prefixGrams,
				 std::vector<unsigned>& candidates);

	bool findCandidatesAndMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);

	bool findCandidatesAndExactMappingsK(unsigned& k, Read& read, std::vector<unsigned>& candidates,
			std::vector<ReadMapping>& results, SingleEndVerifier& verifier);
};

#endif
