// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "globals.h"
#include "gramgen.h"

using namespace std;

const unsigned GramGen::INVALID_GRAM = std::numeric_limits<unsigned>::max();

bool GramGen::nextValidGram(const string& str, unsigned& start, unsigned& end, unsigned& hash) {
	bool found = false;
	do {
		end = start + gramLen;
		if (end > endPos + 1) {
			break;
		}
		found = true;
		hash = 0;
		for (unsigned i = start; i < end; i++) {
			switch (str[i]) {
			case BASE_A:
				hash = hash << 2;
				break;
			case BASE_C:
				hash = (hash << 2) + 1;
				break;
			case BASE_G:
				hash = (hash << 2) + 2;
				break;
			case BASE_T:
				hash = (hash << 2) + 3;
				break;
			default:
				found = false;
				start = i + 1; // skip to next valid char
				break;
			}
			if (!found) // added by Thanh T. Nguyen to get out of the for loop.
			    break;
		}
	} while (!found);
	return found;
}

bool GramGen::firstGram(const string& str, unsigned startPos, unsigned endPos) {
	start = startPos;
	this->endPos = endPos;
	return nextValidGram(str, start, end, hash);
}

bool GramGen::nextGram(const string& str) {
	if (end > endPos) {
		return false;
	}
	switch (str[end]) {
	case BASE_A:
		hash = (hash << 2) & mask;
		break;
	case BASE_C:
		hash = ((hash << 2) + 1) & mask;
		break;
	case BASE_G:
		hash = ((hash << 2) + 2) & mask;
		break;
	case BASE_T:
		hash = ((hash << 2) + 3) & mask;
		break;
	default:
		// Skip to next valid char and init hash.
		start = end + 1;
		if (!nextValidGram(str, start, end, hash))
			return false;
		start--;
		end--;
		break;
	}
	start++;
	end++;
	return true;
}

bool GramGen::generateGrams(const string& str, vector<unsigned>& grams) {
	if (!firstGram(str, 0, str.length() - 1)) {
		return false;
	}
	grams.push_back(hash);
	while (nextGram(str)) {
		grams.push_back(hash);
	}
	return true;
}

bool GramGen::readNextValidGram(const string& str, unsigned& start, unsigned& end, unsigned& hash,
		vector<unsigned>& grams) {
	bool found = false;
	do {
		end = start + gramLen;
		if (end > endPos + 1) {
			break;
		}
		found = true;
		hash = 0;
		for (unsigned i = start; i < end; i++) {
			switch (str[i]) {
			case BASE_A:
				hash = hash << 2;
				break;
			case BASE_C:
				hash = (hash << 2) + 1;
				break;
			case BASE_G:
				hash = (hash << 2) + 2;
				break;
			case BASE_T:
				hash = (hash << 2) + 3;
				break;
			default:
				found = false;
				start++;
				grams.push_back(INVALID_GRAM);
				break;
			}
		}
	} while (!found);
	return found;
}

bool GramGen::readFirstGram(const string& str, unsigned startPos, unsigned endPos,
		vector<unsigned>& grams) {
	start = startPos;
	this->endPos = endPos;
	return readNextValidGram(str, start, end, hash, grams);
}

bool GramGen::readNextGram(const string& str, vector<unsigned>& grams) {
	if (end > endPos) {
		return false;
	}
	switch (str[end]) {
	case BASE_A:
		hash = (hash << 2) & mask;
		break;
	case BASE_C:
		hash = ((hash << 2) + 1) & mask;
		break;
	case BASE_G:
		hash = ((hash << 2) + 2) & mask;
		break;
	case BASE_T:
		hash = ((hash << 2) + 3) & mask;
		break;
	default:
		// Skip to next valid char and init hash.
		start++;
		if (!readNextValidGram(str, start, end, hash, grams)) {
			return false;
		}
		start--;
		end--;
		break;
	}
	start++;
	end++;
	return true;
}

bool GramGen::readGenerateGrams(const string& str, vector<unsigned>& grams) {
	if (!readFirstGram(str, 0, str.length() - 1, grams)) {
		return false;
	}
	grams.push_back(hash);
	while (readNextGram(str, grams)) {
		grams.push_back(hash);
	}
	return true;
}
