// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "index/indexer.h"
#include "index/indexbuilder.h"
#include "thread/threadstarter.h"
#include <vector>
#include <fstream>
#include <iostream>
#include <algorithm>
#include <pthread.h>

using namespace std;

template<>
void
IndexBuilder<BitVecTriple>::
build(bool useProgressIndicator) {
	if(indexPartitions.empty()) {
		cerr << "Need at least one index partition to build index." << endl;
		exit(0);
	}
	sort(indexPartitions.begin(), indexPartitions.end());

	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	unsigned totalElements = 0;
	vector<unsigned> startOffsets;
	for(unsigned i = 0; i < indexPartitions.size(); i++) {
		startOffsets.push_back(totalElements);
		totalElements += (indexPartitions[i].endPos - indexPartitions[i].startPos);
	}

	allElements = new BitVecTriple[totalElements];

	// Start progress indicator thread.
	pthread_t progressIndicatorThreadHandle;
	ProgressIndicator progressIndicator(refSeq.length());
	ProgressIndicator* progIndPtr = NULL;
	if(useProgressIndicator) {
		progIndPtr = &progressIndicator;
	}

	if(useProgressIndicator) {
		if(pthread_create(&progressIndicatorThreadHandle, &attr, startIndexBuildProgressIndicator, &progressIndicator)) {
			cerr << "Fatal error. Failed to create progress indicator thread." << endl;
			exit(0);
		}
	}

	// Start worker threads.
	vector<IndexBuildWorker<BitVecTriple>* > workerPtrs;
	workerPtrs.reserve(indexPartitions.size());
	unsigned partitionsRemaining = indexPartitions.size();
	unsigned partIndex = 0;
	while(partitionsRemaining > 0) {
		unsigned runThreads = min(numThreads, partitionsRemaining);
		pthread_t workerThreadHandles[runThreads];
		for(unsigned i = 0; i < runThreads; i++) {
			IndexPartition& indexPartition = indexPartitions[partIndex];
			IndexBuildWorker<BitVecTriple>* worker = new IndexBuildWorker<BitVecTriple>(refSeq, indexPartition.startPos, indexPartition.endPos, gramLength, allElements + startOffsets[partIndex], progIndPtr);
			workerPtrs.push_back(worker);
			if(pthread_create(&workerThreadHandles[i], &attr, startBitVecIndexBuildWorker, worker)) {
				cerr << "Fatal error. Failed to create mapper thread " << i << "." << endl;
				exit(0);
			}
			partIndex++;
		}
		for(unsigned i = 0; i < runThreads; i++) {
			void* status = 0;
			pthread_join(workerThreadHandles[i], &status);
		}
		partitionsRemaining -= runThreads;
	}

	if(useProgressIndicator) {
		progressIndicator.terminate();
		void* status = 0;
		pthread_join(progressIndicatorThreadHandle, &status);
	}

	mergeToFile(workerPtrs, useProgressIndicator);

	// Cleanup workers.
	for(unsigned i = 0; i < workerPtrs.size(); i++) {
		delete workerPtrs[i];
	}

	delete[] allElements;
}

template<class ElementType>
void
IndexBuilder<ElementType>::
mergeToFile(vector<IndexBuildWorker<BitVecTriple>*>& workerPtrs, bool useProgressIndicator) {

	ofstream out(indexFileName, ios::out | ios::binary);

	// Writing gram length.
	out.write((const char*) &gramLength, sizeof(unsigned));

	// Write dummy numLists to be filled at very end (we will count them during the merge).
	unsigned numLists = 0;
	out.write((const char*)&numLists, sizeof(unsigned));

	// initialize heap and count number of elements for progress indicator.
	unsigned totalElements = 0;
	vector<HeapElement<BitVecTriple> > heap;
	for(unsigned i = 0; i < workerPtrs.size(); i++) {
		// Sanity, is the partition empty?
		if(workerPtrs[i]->begin() != workerPtrs[i]->end()) {
			totalElements += workerPtrs[i]->getElementCount();
			heap.push_back(HeapElement<BitVecTriple>(workerPtrs[i]->begin(), workerPtrs[i]->end()));
		}
	}
	make_heap(heap.begin(), heap.end());

	// Start progress indicator thread if useProgressIndicator == true.
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	// Start progress indicator thread.
	pthread_t progressIndicatorThreadHandle;
	ProgressIndicator progressIndicator(totalElements);
	ProgressIndicator* progIndPtr = NULL;
	if(useProgressIndicator) {
		progIndPtr = &progressIndicator;
	}

	if(useProgressIndicator) {
		if(pthread_create(&progressIndicatorThreadHandle, &attr, startIndexMergeProgressIndicator, &progressIndicator)) {
			cerr << "Fatal error. Failed to create progress indicator thread." << endl;
			exit(0);
		}
	}

	// Merge sorted lists into file.
	while(!heap.empty()) {
		numLists++;
		vector<HeapElement<BitVecTriple> > popped;
		// Pop first element.
		popped.push_back(heap.front());
		pop_heap(heap.begin(), heap.end());
		heap.pop_back();
		// Continue popping elements with same gram.
		if(!heap.empty()) {
			while(heap.front().iter->hash == popped.front().iter->hash) {
				popped.push_back(heap.front());
				pop_heap(heap.begin(), heap.end());
				heap.pop_back();
				if(heap.empty()) break;
			}
		}
		// First determine list size.
		unsigned totalListSize = 0;
		unsigned hash = popped.front().iter->hash;
		for(unsigned i = 0; i < popped.size(); i++) {
			BitVecTriple* iter = popped[i].iter;
			while(iter != popped[i].end && iter->hash == hash) {
				totalListSize++;
				iter++;
			}
		}
		// Write lists.
		out.write((const char*)&hash, sizeof(unsigned));
		out.write((const char*)&totalListSize, sizeof(unsigned));
		for(unsigned i = 0; i < popped.size(); i++) {
			while(popped[i].iter != popped[i].end && popped[i].iter->hash == hash) {
				out.write((const char*)&popped[i].iter->bitVecPair.pos, sizeof(unsigned));
				out.write((const char*)&popped[i].iter->bitVecPair.bitVec, sizeof(BitvectorType));
				popped[i].iter++;
			}
			// Push back to heap (with incremented iterator).
			if(popped[i].iter != popped[i].end) {
				heap.push_back(popped[i]);
				push_heap(heap.begin(), heap.end());
			}
		}
		if(useProgressIndicator) progressIndicator.incrementProgress(totalListSize);
	}
	if(useProgressIndicator) {
		progressIndicator.terminate();
		void* status = 0;
		pthread_join(progressIndicatorThreadHandle, &status);
	}
	// Go back to the place after gram length is written (skip intial unsigned).
	out.seekp(ios_base::beg + sizeof(unsigned));
	out.write((const char*)&numLists, sizeof(unsigned));
	out.close();
}
