// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "globals.h"
#include "index/indexer.h"
#include "index/indexbuilder.h"
#include "tools/statsutil.h"
#include "tools/looptimer.h"

#include <vector>
#include <fstream>
#include <iostream>
#include <string>
#include <cstring>
#include <algorithm>
#include <dirent.h>
#include <tr1/functional>
#include <sys/stat.h>
#include <sys/types.h>

using namespace std;
using namespace tr1;

void Indexer::buildIndex() {
	StatsUtil st;
	st.startTimeMeasurement();

	setGramGen();
	IndexBuilder<BitVecTriple> indexBuilder(*refSeq, indexFileName->c_str(), gramGen.gramLen,
			conf.numThreads);
	unsigned startPos = 0;
	for (unsigned i = 0; i < chromosomeLengths.size(); i++) {
		indexBuilder.addIndexPartition(IndexPartition(startPos, chromosomeLengths[i] - 1));
		startPos = chromosomeLengths[i];
	}
	indexBuilder.build(conf.useProgressIndicator);

	st.stopTimeMeasurement();
	double time = st.getTimeMeasurement(TFSEC);

	clog << "INDEX CONSTRUCTION TIME: " << time << "s" << endl;
	stats.indexingTime = time;
	setIndexStats();
}

void Indexer::readRefSeqFromDir() {
	DIR* directoryp;
	struct dirent* dirFormat;
	struct stat fileStatus;
	vector<string> refFiles;
	const char* dirName = refSeqPath->c_str();
	directoryp = opendir(dirName);
	if (directoryp == NULL) {
		cerr << "Fatal error. Failed to open reference sequence directory: " << *refSeqPath << endl;
		exit(0);
	}
	while ((dirFormat = readdir(directoryp))) {
		string filepath = (*refSeqPath) + "/" + dirFormat->d_name;
		// If the file is a directory (or is in some way invalid) we'll skip it.
		if (stat(filepath.c_str(), &fileStatus)) {
			continue;
		}
		if (S_ISDIR(fileStatus.st_mode)) {
			continue;
		}
		// Check for fasta files only.
		if (strstr(dirFormat->d_name, ".fa")) {
			refFiles.push_back(filepath);
		}
	}
	sort(refFiles.begin(), refFiles.end());
	for (unsigned i = 0; i < refFiles.size(); i++) {
		readRefSeqFromFile(refFiles[i]);
	}
	clog << "Reference sequence files read: " << refFiles.size() << endl;
	closedir(directoryp);
}

void Indexer::readRefSeqFromFile(const string& refSeqFile) {
	const char* fileName = refSeqFile.c_str();
	ifstream in(fileName);
	string line;
	stringstream tempName;
	unsigned i = 0;
	if (!in.is_open()) {
		cerr << "Fatal error. Failed to open reference sequence file: " << refSeqFile << endl;
		exit(0);
	}

	// Reserve the maximum size of ref sequence file.
	// The following save memory about 1G.
	const unsigned PAD = 2048;
	in.seekg(0, ios::end);
	refSeq->reserve(unsigned(in.tellg()) + PAD);
	in.seekg(0, ios::beg);

	clog << "Reading reference sequence from file: " << refSeqFile << endl;
	while (!in.eof()) {
		getline(in, line);
		if (line.size() > 0) {
			if (line[0] != '>') {
				// Convert to uppercase.
				transform(line.begin(), line.end(), line.begin(), (int(*)(int))toupper);
				refSeq->append(line);
			}
			if(line[0] == '>') {
				if(i == 0) {
					tempName<<line.substr(1,line.size());
				} else {
					chromosomeLengths.push_back(refSeq->length());
					chromosomeNames.push_back(tempName.str());
					tempName.str(std::string());
					tempName<<line.substr(1,line.size());
				}
				i++;
			}
		} else {
			chromosomeLengths.push_back(refSeq->length());
			chromosomeNames.push_back(tempName.str());
		}
	}
	in.close();
}

void Indexer::loadIndex() {
	const char* fileName = (*indexFileName).c_str();
	ifstream in(fileName);
	if (!in.is_open()) {
		cerr << "Fatal error. Failed to open index file for reading: " << *indexFileName << endl;
		exit(0);
	}
	// Read gram length from the index.
	in.read((char*) &conf.gramLength, sizeof(unsigned));

	// determine if we use bit vectors
	if((conf.confOptions & ALL_MAPPINGS) && conf.readLength/conf.gramLength > conf.distThresh + 1){
		useBitVector = false;
	}
	else{
		useBitVector = true;
	}

	// Initialize gram generator.
	setGramGen();
	// Read total number of inverted lists.
	unsigned totalInvLists = 0;
	in.read((char*) &totalInvLists, sizeof(unsigned));
	// Preallocate an estimated number of required buckets in the gram hash table.
	unsigned numBucketsEstimate = totalInvLists + (totalInvLists * 0.7f);

	if(useBitVector) bitVecRefInvIndex.rehash(numBucketsEstimate);
	else refInvIndex.rehash(numBucketsEstimate);

	stats.actualIndexMem = 0;
	LoopTimer loopTimer;
	if (conf.useProgressIndicator) {
		loopTimer.begin("LOADING INDEX", totalInvLists);
	}
	// Read inverted lists from file and populate in-memory index.
	for (unsigned i = 0; i < totalInvLists; i++) {
		// Read gram of current list.
		unsigned gramCode;
		in.read((char*) &gramCode, sizeof(unsigned));
		// Read size of current list, and preallocate target vector.
		unsigned currentInvListSize;
		in.read((char*) &currentInvListSize, sizeof(unsigned));
		vector<BitVecPair>* invList = new vector<BitVecPair>();
		invList->reserve(currentInvListSize);
		invList->resize(currentInvListSize);
		// Read contents of inverted list into preallocated vector, and add entry in index.
		in.read((char*) &invList->at(0), sizeof(BitVecPair) * currentInvListSize);

		if(useBitVector){
			bitVecRefInvIndex[gramCode] = invList;
			stats.actualIndexMem += sizeof(BitVecPair) * currentInvListSize;
		}
		else{
			vector<unsigned>* invListNB = new vector<unsigned>();
			invListNB->reserve(currentInvListSize);
			for(unsigned j = 0; j < invList->size(); j++){
				invListNB->push_back((*invList)[j].pos);
			}
			delete invList;

			refInvIndex[gramCode] = invListNB;
			stats.actualIndexMem += sizeof(unsigned) * currentInvListSize;
		}

		if (conf.useProgressIndicator) {
			loopTimer.next();
		}
	}
	if (conf.useProgressIndicator) {
		loopTimer.end();
	}

	if(useBitVector){
		// Rough estimate of memory needed for buckets.
		stats.actualIndexMem += bitVecRefInvIndex.bucket_count() *  sizeof(unsigned);
		// Rough estimate of memory needed for keys.
		stats.actualIndexMem += bitVecRefInvIndex.size() *  sizeof(unsigned);
		// Rough estimate of memory needed for pointer values.
		stats.actualIndexMem += bitVecRefInvIndex.size() * sizeof(vector<BitVecPair>*);
		// Rough estimate of memory needed for vectors values.
		stats.actualIndexMem += bitVecRefInvIndex.size() * sizeof(vector<BitVecPair>);
	}
	else{
		// Rough estimate of memory needed for buckets.
		stats.actualIndexMem += bitVecRefInvIndex.bucket_count() *  sizeof(unsigned);
		// Rough estimate of memory needed for keys.
		stats.actualIndexMem += bitVecRefInvIndex.size() *  sizeof(unsigned);
		// Rough estimate of memory needed for pointer values.
		stats.actualIndexMem += bitVecRefInvIndex.size() * sizeof(vector<unsigned>*);
		// Rough estimate of memory needed for vectors values.
		stats.actualIndexMem += bitVecRefInvIndex.size() * sizeof(vector<unsigned>);
	}

	in.close();
}

void Indexer::addGramToIndex(unsigned hash, unsigned pos) {
	if (bitVecRefInvIndex.find(hash) == bitVecRefInvIndex.end()) {
		bitVecRefInvIndex[hash] = new vector<BitVecPair>();
	}
	BitvectorType align = 0;
	BitvectorType bitVector = generateBitVector(align, *refSeq, pos,
			gramGen.gramLen, refSeq->length());
	BitVecPair tmp(pos, bitVector);
	bitVecRefInvIndex[hash]->push_back(tmp);
}

void Indexer::writeIndexToFile() {
	const char* fileName = indexFileName->c_str();
	ofstream out(fileName);
	if (!out.is_open()) {
		cerr << "Fatal error. Failed to open index file for writing: " << *indexFileName << endl;
		exit(0);
	}
	// Write gram length.
	unsigned gramLen = conf.gramLength;
	out.write((const char*) &gramLen, sizeof(unsigned));
	// Write total number of inverted lists.
	unsigned indexSize = bitVecRefInvIndex.size();
	out.write((const char*) &indexSize, sizeof(unsigned));
	// Write inverted lists.
	for (BitVectorInvertedIndex::iterator it = bitVecRefInvIndex.begin();
			it != bitVecRefInvIndex.end(); it++) {
		// Write gram of current list.
		out.write((const char*) &it->first, sizeof(unsigned));
		// Write size of current list.
		unsigned invListSize = it->second->size();
		out.write((const char*) &invListSize, sizeof(unsigned));
		// Write contents of current list.
		out.write((const char*) &it->second->at(0), sizeof(BitVecPair) * invListSize);
	}
	out.close();
}

void Indexer::setIndexStats() {
	stats.refSeqLen = refSeq->length();
	stats.totalIndexSize = 0;
	stats.actualIndexMem = 0;
	for (BitVectorInvertedIndex::iterator it = bitVecRefInvIndex.begin();
			it != bitVecRefInvIndex.end(); it++) {
		stats.totalIndexSize += it->second->size();
		stats.actualIndexMem += it->second->capacity();
	}
	stats.totalIndexSize *= sizeof(BitVecPair);
	stats.actualIndexMem *= sizeof(BitVecPair);
}
