// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "input/readinfo.h"
#include "output/resultstringwriter.h"
#include "index/indexer.h"
#include "search/searcher.h"
#include "search/singleendsearcher.h"
#include "search/pairedendsearcher.h"

#include <algorithm>
#include <iostream>
#include <boost/lexical_cast.hpp>

using namespace std;

void ResultStringWriter::getChromosomeOffset(vector<string>& chrNames, vector<unsigned>& chrLengths,
		unsigned absolutePos, string& chrName, unsigned& relativePos) const {
	chrName.clear();
	relativePos = 0;
	vector<unsigned>::iterator itr = upper_bound(chrLengths.begin(), chrLengths.end(), absolutePos);
	if (*itr > chrLengths[0]) {
		relativePos = (absolutePos - *(itr - 1));
	} else {
		relativePos = absolutePos;
	}
	chrName = chrNames[(itr - chrLengths.begin())];
}

void SAMFormatResultStringWriter::createInitialString(stringstream& resultStream,
		Indexer& indexer) const {
	if (conf.samOutputOptions == SAM_NO_HEADERS) {
		return;
	}
	// Writing HD TAG.
	resultStream << HD_TAG << SEPARATOR << VERSION_TAG << SAM_VERSION << SEPARATOR << SO_TAG
			<< "unsorted" << endl;
	if (conf.samOutputOptions == SAM_NO_SEQ) {
		resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR
		     << PROGNAME_TAG << SEPARATOR << CMD_LINE_TAG << "\"" << conf.commandLineInput << "\"" << endl;
		return;
	}
	// Writing SQ TAGS.
	unsigned currLen = 0;
	unsigned prevLen = 0;
	for (unsigned i = 0; i < indexer.chromosomeLengths.size(); i++) {
		currLen = indexer.chromosomeLengths[i];
		resultStream << SEQ_TAG << SEPARATOR << SEQNAME_TAG << indexer.chromosomeNames[i]
				<< SEPARATOR << SEQLEN_TAG << (currLen - prevLen) << endl;
		prevLen = currLen;
	}
	// Writing PROG TAG.
	resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR
		     << PROGNAME_TAG << SEPARATOR << CMD_LINE_TAG << "\"" << conf.commandLineInput << "\"" << endl;
}

void SAMFormatResultStringWriter::writeResultString(stringstream& resultStream, ReadInfo& readInfo,
		Searcher* searcher) {
	SingleEndSearcher* concreteSearcher = static_cast<SingleEndSearcher*>(searcher);
	vector<ReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<ReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<ReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<ReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	unsigned relativePos = 0;
	unsigned samFlags = 0;

	vector<ReadMapping>::iterator begin;
	for(begin = forwardIter; forwardIter != forwardEnd; forwardIter++){
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);

		// For 1-based offset.
		relativePos++;

		if(forwardIter != begin) samFlags = NOT_PRIMARY_ALIGN;

		resultStream << readInfo.readId << SEPARATOR << samFlags << SEPARATOR << chrName
				<< SEPARATOR << relativePos << SEPARATOR;
		resultStream << mapQuality << SEPARATOR;
		resultStream << forwardIter->cigar << SEPARATOR;
		resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
		if(forwardIter == begin)
			resultStream << templateLen << SEPARATOR << readInfo.readString << SEPARATOR
				<< readInfo.readQuality << SEPARATOR;
		else	
			resultStream << templateLen << SEPARATOR << "*" << SEPARATOR << "*" << SEPARATOR;
		resultStream << NM_TAG << forwardIter->distance << SEPARATOR << XM_TAG
				<< forwardIter->distance;
		resultStream << endl;
	} 

	for(begin = reverseIter; reverseIter != reverseEnd; reverseIter++){
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->pos, chrName, relativePos);

		relativePos++;

		if(reverseIter == begin) samFlags = READ_REVERSE_MAPPED;
		else samFlags = READ_REVERSE_MAPPED + NOT_PRIMARY_ALIGN;

		resultStream << readInfo.readId << SEPARATOR << samFlags << SEPARATOR << chrName
				<< SEPARATOR << relativePos << SEPARATOR;
		resultStream << mapQuality << SEPARATOR;

		resultStream << reverseIter->cigar << SEPARATOR;
		resultStream << refNext << SEPARATOR << posNext << SEPARATOR;
		if(reverseIter == begin)
			resultStream << templateLen << SEPARATOR << readInfo.reverseReadString << SEPARATOR
					<< readInfo.reverseReadQuality << SEPARATOR;
		else
			resultStream << templateLen << SEPARATOR << "*" << SEPARATOR << "*" << SEPARATOR;

		resultStream << NM_TAG << reverseIter->distance << SEPARATOR << XM_TAG
				<< reverseIter->distance;
		resultStream << endl;
	}
}

void PairedSAMFormatResultStringWriter::createInitialString(stringstream& resultStream,
		Indexer& indexer) const {
	if (conf.samOutputOptions == SAM_NO_HEADERS) {
		return;
	}
	// Writing HD TAG.
	resultStream << HD_TAG << SEPARATOR << VERSION_TAG << SAM_VERSION << SEPARATOR << SO_TAG
			<< "unsorted" << endl;
	if (conf.samOutputOptions == SAM_NO_SEQ) {
		resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR
			<< PROGNAME_TAG << SEPARATOR << CMD_LINE_TAG << "\"" << conf.commandLineInput << "\"" << endl;
		return;
	}
	// Writing SQ TAGS.
	unsigned currLen = 0;
	unsigned prevLen = 0;
	for (unsigned i = 0; i < indexer.chromosomeLengths.size(); i++) {
		currLen = indexer.chromosomeLengths[i];
		resultStream << SEQ_TAG << SEPARATOR << SEQNAME_TAG << indexer.chromosomeNames[i]
				<< SEPARATOR << SEQLEN_TAG << (currLen - prevLen) << endl;
		prevLen = currLen;
	}
	// Writing PROG TAG.
	resultStream << PROG_TAG << SEPARATOR << VERSION_TAG << HOBBES_VERSION << SEPARATOR
			<< PROGNAME_TAG << SEPARATOR << CMD_LINE_TAG << "\"" << conf.commandLineInput << "\"" << endl;
}

void PairedSAMFormatResultStringWriter::writeResultString(stringstream& resultStream,
		ReadInfo& readInfo, Searcher* searcher) {
	PairedEndSearcher* concreteSearcher = static_cast<PairedEndSearcher*>(searcher);
	PairedEndReadInfo* concreteReadSt = static_cast<PairedEndReadInfo*>(&readInfo);
	vector<PairedReadMapping>::iterator forwardIter = concreteSearcher->forwardResults.begin();
	vector<PairedReadMapping>::iterator forwardEnd = concreteSearcher->forwardResults.end();
	vector<PairedReadMapping>::iterator reverseIter = concreteSearcher->reverseResults.begin();
	vector<PairedReadMapping>::iterator reverseEnd = concreteSearcher->reverseResults.end();
	string chrName;
	string chrName2;
	unsigned relativePos = 0;
	unsigned relativePos2 = 0;
	char refNext = '='; //unavailable

	unsigned samFlags = 0;

	vector<PairedReadMapping>::iterator begin;
	for(begin = forwardIter; forwardIter != forwardEnd; forwardIter++) {
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->pos, chrName, relativePos);
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, forwardIter->posRight, chrName2,
				relativePos2);

		// For 1-based offset.
		relativePos++;
		relativePos2++;
		int nextPos = relativePos2 - relativePos + concreteReadSt->first.readString.size();

		if(forwardIter == begin) samFlags = READ_PAIR_FORWARD_MAPPED1;
		else samFlags = READ_PAIR_FORWARD_MAPPED1 + NOT_PRIMARY_ALIGN;

		resultStream << concreteReadSt->first.readId << SEPARATOR;
		resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
				<< SEPARATOR;
		resultStream << mapQuality << SEPARATOR;

		resultStream << forwardIter->cigar << SEPARATOR;
		resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR << nextPos << SEPARATOR;
		if(forwardIter == begin)
			resultStream << concreteReadSt->first.readString << SEPARATOR
					<< concreteReadSt->first.readQuality << SEPARATOR;
		else
			resultStream << "*" << SEPARATOR << "*" << SEPARATOR;

		resultStream << NM_TAG << forwardIter->distance << endl;

		if(forwardIter == begin) samFlags = READ_PAIR_FORWARD_MAPPED2;
		else samFlags = READ_PAIR_FORWARD_MAPPED2 + NOT_PRIMARY_ALIGN;

		int nextPos2 = -nextPos;
		resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
				<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
		resultStream << mapQuality << SEPARATOR;

		resultStream << forwardIter->cigarRight << SEPARATOR;
		resultStream << refNext << SEPARATOR << relativePos << SEPARATOR << nextPos2 << SEPARATOR;

		if(forwardIter == begin)
			resultStream << concreteReadSt->second.readString << SEPARATOR
					<< concreteReadSt->second.readQuality << SEPARATOR;
		else
			resultStream << "*" << SEPARATOR << "*" << SEPARATOR;
		resultStream << NM_TAG << forwardIter->distanceRight << endl;
	}

	for(begin = reverseIter; reverseIter != reverseEnd; reverseIter++) {
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->posRight, chrName, relativePos);
		getChromosomeOffset(searcher->indexer->chromosomeNames,
				searcher->indexer->chromosomeLengths, reverseIter->pos, chrName2, relativePos2);

		relativePos++;
		relativePos2++;
		int nextPos = relativePos2 - relativePos - concreteReadSt->first.readString.size();

		if(reverseIter == begin) samFlags = READ_PAIR_REVERSE_MAPPED1;
		else samFlags = READ_PAIR_REVERSE_MAPPED1 + NOT_PRIMARY_ALIGN;

		resultStream << concreteReadSt->first.readId << SEPARATOR;
		resultStream << samFlags << SEPARATOR << chrName << SEPARATOR << relativePos
				<< SEPARATOR;
		resultStream << mapQuality << SEPARATOR;

		resultStream << reverseIter->cigarRight << SEPARATOR;
		resultStream << refNext << SEPARATOR << relativePos2 << SEPARATOR << nextPos << SEPARATOR;
		if(reverseIter == begin)
			resultStream << concreteReadSt->second.reverseReadString
					<< SEPARATOR << concreteReadSt->second.reverseReadQuality << SEPARATOR;
		else
			resultStream << "*" << SEPARATOR << "*" << SEPARATOR;
		resultStream << NM_TAG << reverseIter->distanceRight << endl;

		if(reverseIter == begin) samFlags = READ_PAIR_REVERSE_MAPPED2;
		else samFlags = READ_PAIR_REVERSE_MAPPED2 + NOT_PRIMARY_ALIGN;

		int nextPos2 = -nextPos;
		resultStream << concreteReadSt->second.readId << SEPARATOR << samFlags << SEPARATOR
				<< chrName << SEPARATOR << relativePos2 << SEPARATOR;
		resultStream << mapQuality << SEPARATOR;

		resultStream << reverseIter->cigar << SEPARATOR;
		resultStream << refNext << SEPARATOR << relativePos << SEPARATOR << nextPos2 << SEPARATOR;
		if(reverseIter == begin)
			resultStream << concreteReadSt->first.reverseReadString
					<< SEPARATOR << concreteReadSt->first.reverseReadQuality << SEPARATOR;
		else
			resultStream << "*" << SEPARATOR << "*" << SEPARATOR;
		resultStream << NM_TAG << reverseIter->distance << endl;
	}
}
