// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "search/pairedendsearcher.h"

#include <cstring>
#include <iostream>
#include <cstdlib>

using namespace std;

MappingResultType PairedEndSearcher::getMappingResultType() {
	if (forwardResults.size() == 0 && reverseResults.size() == 0) {
		return MRT_UNMAPPED;
	} else {
		return MRT_MAPPED;
	}
}

void PairedEndSearcher::prepare(unsigned readLen) {
	forwardCandidates1.clear();
	forwardCandidates2.clear();
	forwardResults.clear();
	reverseCandidates1.clear();
	reverseCandidates2.clear();
	reverseResults.clear();
	verifier->setMinMaxDist(readLen);

	verifier->updateStats = true;
	memset(verifier->tempIndel, 0, (2 * distThresh + 1) * sizeof(long));
	memset(verifier->tempMisMatch, 0, (2 * distThresh + 1) * sizeof(long));
	memset(verifier->singleEndVerifier->tempIndel, 0, (2 * distThresh + 1) * sizeof(long));
	memset(verifier->singleEndVerifier->tempMisMatch, 0, (2 * distThresh + 1) * sizeof(long));
}

void PairedEndSearcher::getStats(std::pair<unsigned, unsigned>& counts) {
	counts.first = forwardResults.size() + reverseResults.size();
	counts.second = forwardCandidates1.size() + reverseCandidates1.size()
			+ forwardCandidates2.size() + reverseCandidates2.size();
}

void PairedEndSearcher::getBestStats() {
	if(forwardResults.empty() && reverseResults.empty()) return;

	if (verifier->tempMisMatch[0] > 0) {
		mismatches[0]++;
	} else {
		for (unsigned j = 1; j <= 2 * distThresh; j++) {
			if (verifier->tempMisMatch[j] != 0 || verifier->tempIndel[j] != 0) {
				if (verifier->tempMisMatch[j] >= verifier->tempIndel[j]) {
					mismatches[j]++;
				} else{
					indels[j]++;
				}
				break;
			}
		}
	}
}

void PairedEndSearcher::findAllMappings(ReadInfo& readInfo) {

	if(conf.maxMappingsPerRead != 0){
		cerr << "Error: m mapping mode not supported for paired-end reads." << endl;
		exit(0);
	}

	PairedEndReadInfo* pairedEndReadInfo = static_cast<PairedEndReadInfo*>(&readInfo);
	setPairedRead(*pairedEndReadInfo);
	unsigned tOccurThresh = 0;
    // Always use both forward and reverse for paired-end mapping.
	forwardRead.setReadStringFromReadInfo(*pairedEndReadInfo);
	if (fillRead(forwardRead, tOccurThresh)) {
		mergeAndVerifyAll(forwardRead, forwardCandidates1, forwardCandidates2, forwardResults);
	}
	genReverses(*pairedEndReadInfo);
	reverseRead.setReverseReadStringFromReadInfo(*pairedEndReadInfo);
	if (fillRead(reverseRead, tOccurThresh)) {
		mergeAndVerifyAll(reverseRead, reverseCandidates1, reverseCandidates2, reverseResults);
	}
}

void PairedEndSearcher::findKMappings(ReadInfo& readInfo) {
	cerr << "Error: K mappings mode not supported for paired-end reads." << endl;
	exit(0);
}

void PairedEndSearcher::mergeAndVerifyAll(PairedEndRead& readStructure, vector< unsigned>& candidates1, vector< unsigned>& candidates2, vector<PairedReadMapping>& results){
	merger->findCandidatesAll(readStructure, candidates1, candidates2);
	verifier->verifyCandidates(readStructure, candidates1, candidates2, results, conf.maxMappingsPerRead);
}

bool PairedEndSearcher::fillRead(PairedEndRead& read, unsigned& tOccurThresh) {
	return (Searcher::fillRead(read.first, tOccurThresh)
			&& Searcher::fillRead(read.second, tOccurThresh));
}

void PairedEndSearcher::setPairedRead(PairedEndReadInfo& readInfo){
	getReverseComplementRead(readInfo.second.readString);
	getReverseReadQuality(readInfo.second.readQuality);
}

void PairedEndSearcher::genReverses(PairedEndReadInfo& readInfo) {
	readInfo.setReverseReadString(readInfo.readString);
	getReverseComplementRead(readInfo.reverseReadString);
	readInfo.setReverseReadQuality(readInfo.readQuality);
	getReverseReadQuality(readInfo.reverseReadQuality);
	readInfo.first.setReverseReadString(readInfo.first.readString);
	readInfo.second.setReverseReadString(readInfo.second.readString);
	getReverseComplementRead(readInfo.second.reverseReadString);
	getReverseComplementRead(readInfo.first.reverseReadString);
	readInfo.first.setReverseReadQuality(readInfo.first.readQuality);
	readInfo.second.setReverseReadQuality(readInfo.second.readQuality);
	getReverseReadQuality(readInfo.second.reverseReadQuality);
	getReverseReadQuality(readInfo.first.reverseReadQuality);
	string tmp = readInfo.first.reverseReadString;
	readInfo.first.setReverseReadString(readInfo.second.reverseReadString);
	readInfo.second.setReverseReadString(tmp);
	tmp = readInfo.first.reverseReadQuality;
	readInfo.first.setReverseReadQuality(readInfo.second.reverseReadQuality);
	readInfo.second.setReverseReadQuality(tmp);
}
