// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "include/search/read.h"
#include "include/bitvectorfilter.h"
#include <algorithm>
#include <limits>

using namespace std;

vector<BitVecPair> Read::bitvectorNullList;
vector<unsigned> Read::nullList;

void Read::getInvertedLists(Indexer* indexer, GramGen& gramGen) {
	grams.clear();
	gramGen.readGenerateGrams(*readStr, grams);

	if(indexer->useBitVector){
		getInvertedLists(&indexer->bitVecRefInvIndex);
	} else{
		getInvertedLists(&indexer->refInvIndex);
	}
}

void Read::getInvertedLists(BitVectorInvertedIndex* invIndex) {
	invertedLists.clear();

	bitvectorInvertedLists.clear();
	for (unsigned i = 0; i < grams.size(); i++) {
		BitVectorInvertedIndex::iterator itr = invIndex->find(grams[i]);
		if (itr != invIndex->end()) {
			bitvectorInvertedLists.push_back(make_pair(i, itr->second));
		} else {
			if (grams[i] != GramGen::INVALID_GRAM) {
				bitvectorInvertedLists.push_back(make_pair(i, &bitvectorNullList));
			}
		}
	}
}

void Read::getInvertedLists(InvertedIndex* invIndex) {
	bitvectorInvertedLists.clear();

	invertedLists.clear();
	for (unsigned i = 0; i < grams.size(); i++) {
		InvertedIndex::iterator itr = invIndex->find(grams[i]);
		if (itr != invIndex->end()) {
			invertedLists.push_back(make_pair(i, itr->second));
		} else {
			if (grams[i] != GramGen::INVALID_GRAM) {
				invertedLists.push_back(make_pair(i, &nullList));
			}
		}
	}
}

void Read::getPrefixDynamicProgrammingWithBitVector(unsigned distThresh, unsigned seedLen, unsigned gramLen)
{
	// to retrieve 'distThresh + additionalPrefix' prefix lists
	// note that additionalPrefix is either 0 or 1
	distThresh += additionalPrefix;

	prefixGrams.clear();
	if (distThresh == 0) {
		prefixLen = 1;
		unsigned minPrefix = 0;
		unsigned minSize = bitvectorInvertedLists[0].second->size();
		for (unsigned i = 1; i < grams.size(); i++) {
			if (bitvectorInvertedLists[i].second->size() < minSize) {
				minSize = bitvectorInvertedLists[i].second->size();
				minPrefix = i;
			}
		}
		prefixGrams.push_back(minPrefix);
		return;
	}

	unsigned edGramL = distThresh * gramLen;
	unsigned L = seedLen - gramLen + 1;

	if (seedLen < (distThresh - additionalPrefix + 1) * gramLen){
		additionalPrefix = 0; // additionalPrefix is already zero. just make it concrete.

		prefixGrams.reserve(L - edGramL);
		for (unsigned i = 0; i < bitvectorInvertedLists.size(); i++) {
			if (bitvectorInvertedLists[i].second != &bitvectorNullList) {
				prefixGrams.push_back(i);
			}
		}
		return;
	}

	prefixGrams.reserve(distThresh + 1);
	vector<unsigned> size;
	size.reserve(L);
	unsigned j = 0;
	for (unsigned i = 0; i < L; i++) {
		if (j >= bitvectorInvertedLists.size() || bitvectorInvertedLists[j].first != i) {
			size[i] = -1;
		} else {
			size[i] = bitvectorInvertedLists[j].second->size();
			j++;
		}
	}

	unsigned S[L - edGramL];
	unsigned index[distThresh + 1][L - edGramL];
	S[0] = size[0];
	index[0][0] = 0;
	for (unsigned j = 1; j <= L - edGramL; j++) {
		if (size[j] < S[j - 1]) {
			S[j] = size[j];
			index[0][j] = j;
		} else {
			S[j] = S[j - 1];
			index[0][j] = index[0][j - 1];
		}
	}

	for (unsigned i = 1; i <= distThresh; i++) {
		unsigned igramL = i * gramLen;
		if (size[igramL] == (unsigned) -1 || S[0] == (unsigned) -1) {
			S[0] = -1;
		} else {
			S[0] += size[igramL];
		}
		index[i][0] = 0;
		for (unsigned j = 1; j < L - edGramL; j++) {
			if (size[j + igramL] == (unsigned) -1 || size[j + igramL] + S[j] >= S[j - 1]
					|| S[j] == (unsigned) -1) {
				S[j] = S[j - 1];
				index[i][j] = index[i][j - 1];
			} else {
				S[j] += size[j + igramL];
				index[i][j] = j;
			}
		}
	}

	if (S[L - edGramL - 1] == (unsigned) -1) {
		if(additionalPrefix != 0){ // recalculate prefix lists
			distThresh -= additionalPrefix;
			additionalPrefix -= 1;
			getPrefixDynamicProgrammingWithBitVector(distThresh, seedLen, gramLen);
		}
		else{
			prefixGrams.reserve(L - edGramL);
			for (unsigned i = 0; i < bitvectorInvertedLists.size(); i++) {
				if (bitvectorInvertedLists[i].second != &bitvectorNullList) {
					prefixGrams.push_back(i);
				}
			}
		}
		return;
	}
	unsigned t = distThresh;
	unsigned y = L - edGramL - 1;
	unsigned bitVectorPtr = bitvectorInvertedLists.size() - 1;
	for (int i = distThresh; i >= 0; i--) {
		y = index[t][y];
		while (bitvectorInvertedLists[bitVectorPtr].first != (y + i * gramLen)) {
			bitVectorPtr--;
		}
		prefixGrams.push_back(bitVectorPtr);
		t--;
	}
	for (unsigned i = 0; i < prefixGrams.size(); i++) {
		for (unsigned j = i + 1; j < prefixGrams.size(); j++) {
			if (bitvectorInvertedLists[prefixGrams[i]].second->size()
					> bitvectorInvertedLists[prefixGrams[j]].second->size()) {
				unsigned tmp = prefixGrams[i];
				prefixGrams[i] = prefixGrams[j];
				prefixGrams[j] = tmp;
			}
		}
	}
	prefixLen = distThresh + 1;
}

// optimal prefix selection with normal inverted lists
void Read::getPrefixDynamicProgrammingWithoutBitVector(unsigned distThresh, unsigned seedLen, unsigned gramLen) 
{
	distThresh += additionalPrefix;

	prefixGrams.clear();
	if (distThresh == 0) {
		prefixLen = 1;
		unsigned minPrefix = 0;
		unsigned minSize = invertedLists[0].second->size();
		for (unsigned i = 1; i < grams.size(); i++) {
			if (invertedLists[i].second->size() < minSize) {
				minSize = invertedLists[i].second->size();
				minPrefix = i;
			}
		}
		prefixGrams.push_back(minPrefix);
		return;
	}

	unsigned edGramL = distThresh * gramLen;
	unsigned L = seedLen - gramLen + 1;

	if (seedLen < (distThresh - additionalPrefix + 1) * gramLen){
		additionalPrefix = 0; // additionalPrefix is already zero. just make it concrete.

		prefixGrams.reserve(L - edGramL);
		for (unsigned i = 0; i < invertedLists.size(); i++) {
			if (invertedLists[i].second != &nullList) {
				prefixGrams.push_back(i);
			}
		}
		return;
	}

	prefixGrams.reserve(distThresh + 1);
	vector<unsigned> size;
	size.reserve(L);
	unsigned j = 0;
	for (unsigned i = 0; i < L; i++) {
		if (j >= invertedLists.size() || invertedLists[j].first != i) {
			size[i] = -1;
		} else {
			size[i] = invertedLists[j].second->size();
			j++;
		}
	}
	unsigned S[L - edGramL];
	unsigned index[distThresh + 1][L - edGramL];
	S[0] = size[0];
	index[0][0] = 0;
	for (unsigned j = 1; j <= L - edGramL; j++) {
		if (size[j] < S[j - 1]) {
			S[j] = size[j];
			index[0][j] = j;
		} else {
			S[j] = S[j - 1];
			index[0][j] = index[0][j - 1];
		}
	}
	for (unsigned i = 1; i <= distThresh; i++) {
		unsigned igramL = i * gramLen;
		if (size[igramL] == (unsigned) -1 || S[0] == (unsigned) -1) {
			S[0] = -1;
		} else {
			S[0] += size[igramL];
		}
		index[i][0] = 0;
		for (unsigned j = 1; j < L - edGramL; j++) {
			if (size[j + igramL] == (unsigned) -1 || size[j + igramL] + S[j] >= S[j - 1]
					|| S[j] == (unsigned) -1) {
				S[j] = S[j - 1];
				index[i][j] = index[i][j - 1];
			} else {
				S[j] += size[j + igramL];
				index[i][j] = j;
			}
		}
	}
	if (S[L - edGramL - 1] == (unsigned) -1) {
		if(additionalPrefix != 0){
			distThresh -= additionalPrefix;
			additionalPrefix -= 1;
			getPrefixDynamicProgrammingWithoutBitVector(distThresh, seedLen, gramLen);
		}
		else{
			prefixGrams.reserve(L - edGramL);
			for (unsigned i = 0; i < invertedLists.size(); i++) {
				if (invertedLists[i].second != &nullList) {
					prefixGrams.push_back(i);
				}
			}
		}
		return;
	}
	unsigned t = distThresh;
	unsigned y = L - edGramL - 1;
	unsigned vectorPtr = invertedLists.size() - 1;
	for (int i = distThresh; i >= 0; i--) {
		y = index[t][y];
		while (invertedLists[vectorPtr].first != (y + i * gramLen)) {
			vectorPtr--;
		}
		prefixGrams.push_back(vectorPtr);
		t--;
	}
	for (unsigned i = 0; i < prefixGrams.size(); i++) {
		for (unsigned j = i + 1; j < prefixGrams.size(); j++) {
			if (invertedLists[prefixGrams[i]].second->size() > invertedLists[prefixGrams[j]].second->size()) {
				unsigned tmp = prefixGrams[i];
				prefixGrams[i] = prefixGrams[j];
				prefixGrams[j] = tmp;
			}
		}
	}
	prefixLen = distThresh + 1;
}

void PairedEndRead::setReadStringFromReadInfo(PairedEndReadInfo& inputReadStruct) {
	this->first.readStr = &inputReadStruct.first.readString;
	this->second.readStr = &inputReadStruct.second.readString;
}

void PairedEndRead::setReverseReadStringFromReadInfo(PairedEndReadInfo& inputReadStruct) {
	this->first.readStr = &inputReadStruct.first.reverseReadString;
	this->second.readStr = &inputReadStruct.second.reverseReadString;
}

void Read::getEdJoinPrefix(unsigned distThresh, unsigned tOccurThresh, unsigned gramLen) {
	prefixGrams.clear();
	if (distThresh == 0) {
		prefixLen = 1;
		prefixGrams.push_back(0);
		return;
	}
	unsigned left = distThresh + 1;
	unsigned right = 0;
	right = bitvectorInvertedLists.size() - tOccurThresh + 1;
	vector<unsigned> offsets;
	offsets.reserve(right);
	unsigned mid = (left + right) / 2;
	for (unsigned i = 0; i <= mid; i++) {
		offsets.push_back(bitvectorInvertedLists[i].first);
	}
	unsigned err;
	while (left < right) {
		err = getMinEditErrors(offsets, gramLen);
		if (err <= distThresh) {
			left = mid + 1;
			mid = (left + right) / 2;
			for (unsigned i = left; i <= mid; i++) {
				offsets.push_back(bitvectorInvertedLists[i].first);
			}
		} else {
			right = mid;
			mid = (left + right) / 2;
			for (unsigned i = right; i > mid; i--) {
				offsets.pop_back();
			}
		}
	}
	vector<unsigned> prefixOffstes;
	prefixOffstes.reserve(prefixLen);
	getPrefixOffsets(offsets, gramLen, prefixOffstes);
	if (prefixOffstes.size() <= distThresh) {
		prefixLen = bitvectorInvertedLists.size() - tOccurThresh + 1;
		prefixGrams.reserve(prefixLen);
		for (unsigned i = 0; i < prefixLen; i++) {
			prefixGrams.push_back(i);
		}
		return;
	}
	prefixLen = left + 1;
	prefixGrams.reserve(prefixLen);
	for (unsigned i = 0; i < prefixLen; i++) {
		for (unsigned j = 0; j < prefixOffstes.size(); j++) {
			unsigned temp = 0;
			temp = bitvectorInvertedLists[i].first;
			if (prefixOffstes.at(j) == temp) {
				prefixGrams.push_back(i);
			}
		}
	}
}

unsigned Read::getMinEditErrors(vector<unsigned> offsets, unsigned gramLength) {
	sort(offsets.begin(), offsets.end());
	unsigned count = 1;
	unsigned loc = offsets[0] + gramLength - 1;
	for (unsigned i = 0; i < offsets.size(); i++)
		if (offsets[i] > loc) {
			count++;
			loc = offsets[i] + gramLength - 1;
		}
	return count;
}

unsigned Read::getPrefixOffsets(vector<unsigned>& offsets, unsigned gramLen,
		vector<unsigned>& prefixOffsets) {
	sort(offsets.begin(), offsets.end());
	unsigned count = 1;
	unsigned loc = offsets[0] + gramLen - 1;
	prefixOffsets.push_back(offsets[0]);
	for (unsigned i = 0; i < offsets.size(); i++)
		if (offsets[i] > loc) {
			count++;
			prefixOffsets.push_back(offsets[i]);
			loc = offsets[i] + gramLen - 1;
		}
	return count;
}

void Read::getPrefixRecursive(unsigned distThresh, unsigned tOccurThresh, unsigned gramLen) {
	prefixGrams.clear();
	if (distThresh == 0) {
		prefixLen = 1;
		prefixGrams.push_back(0);
		return;
	}
	prefixGrams.reserve(distThresh + 1);
	unsigned minSum = numeric_limits<unsigned>::max();
	vector<unsigned> tmp;
	getPrefixRecursive(-gramLen, distThresh + 1, 0, minSum, prefixGrams, tmp, gramLen);
	for (unsigned i = 0; i < prefixGrams.size(); i++) {
		for (unsigned j = i + 1; j < prefixGrams.size(); j++) {
			if (bitvectorInvertedLists[prefixGrams[i]].second->size()
					> bitvectorInvertedLists[prefixGrams[j]].second->size()) {
				unsigned tmp = prefixGrams[i];
				prefixGrams[i] = prefixGrams[j];
				prefixGrams[j] = tmp;
			}
		}
	}
}

void Read::getPrefixRecursive(int lastPos, unsigned distRemain, unsigned partialSum,
		unsigned &minSum, vector<unsigned>& minPath, vector<unsigned> path, unsigned gramLen) {
	if (distRemain == 0) {
		if (partialSum < minSum) {
			minSum = partialSum;
			minPath = path;
		}
		return;
	}
	distRemain--;
	unsigned start = lastPos + gramLen;
	unsigned end = bitvectorInvertedLists.size() - distRemain * gramLen;
	for (unsigned i = start; i < end; i++) {
		path.push_back(i);
		getPrefixRecursive(i, distRemain, partialSum + bitvectorInvertedLists[i].second->size(),
				minSum, minPath, path, gramLen);
		path.pop_back();
	}
}
