// ============================================================================
//                    Hobbes Genome Sequence Mapping
// ============================================================================
// Copyright (c) 2011-2012, The Regents of the University of California, Irvine
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of The University of California nor the names of
//       its contributors may be used to endorse or promote products derived
//       from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL KNUT REINERT OR THE FU BERLIN BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
// DAMAGE.
//
// ============================================================================

#include "tools/looptimer.h"

#include <iostream>
#include <cstdio>

using namespace std;

LoopTimer::LoopTimer() {
	file = stderr;
	running = false;
}

void LoopTimer::print() {
	if (desc.empty()){
		return;
	}
	if(count != 0)
		fprintf(file, "\r%u%% %s: %u/%u; ", (unsigned) ((double) step * (double) 100 / (double) count),
				desc.c_str(), step, count);
	else
		fprintf(file, "\r*%% %s: %u/*; ", desc.c_str(), step);

	double t = totalTime();
	t /= 1000.0;
	fprintf(file, "%u'%u\"/", (unsigned) (t / 60.0), (unsigned) t % 60);
	if (step && count != 0){
		t = t * (double) count / (double) step;
		fprintf(file, "%u'%u\"   ", (unsigned) (t / 60.0), (unsigned) t % 60);
	} else {
		fprintf(file, "*   ");
	}
}

void LoopTimer::begin(const string &desc, unsigned count) {
	this->desc = desc;
	this->count = count;
	this->step = 0;
	sumTime.tv_sec = 0;
	sumTime.tv_nsec = 0;
	running = true;
	clock_gettime(CLOCK_REALTIME, &startTime);
	currTime = startTime;
	prevTime = currTime;
	print();
}

void LoopTimer::next(unsigned step, unsigned count) {
	this->step += step;
	if (count) {
		this->count = count + this->step;
	}
	if (!running) {
		return;
	}
	clock_gettime(CLOCK_REALTIME, &currTime);
	if ((currTime.tv_sec - prevTime.tv_sec) * 1000 + (currTime.tv_nsec - prevTime.tv_nsec) / 1000000
			>= 200) {
		prevTime = currTime;
		print();
	}
}

void LoopTimer::end(unsigned max) {
	if(max != 0) count = max;
	step = count;
	pause();
	print();
	if (!desc.empty() && !count == 0) {
		fprintf(file, "\n");
	}
}

void LoopTimer::pause() {
	if (!running) {
		return;
	}
	running = false;
	clock_gettime(CLOCK_REALTIME, &currTime);
	sumTime.tv_sec += currTime.tv_sec - startTime.tv_sec;
	sumTime.tv_nsec += currTime.tv_nsec - startTime.tv_nsec;
	if (sumTime.tv_nsec < 0) {
		sumTime.tv_nsec += 1000000000;
		sumTime.tv_sec--;
	} else if (sumTime.tv_nsec >= 1000000000) {
		sumTime.tv_nsec -= 1000000000;
		sumTime.tv_sec++;
	}
	startTime = currTime;
}

void LoopTimer::resume() {
	if (running) {
		return;
	}
	running = true;
	clock_gettime(CLOCK_REALTIME, &startTime);
	currTime = startTime;
}

double LoopTimer::totalTime() {
	return (currTime.tv_sec - startTime.tv_sec + sumTime.tv_sec) * 1000.0
			+ (currTime.tv_nsec - startTime.tv_nsec + sumTime.tv_nsec) / 1000000.0;
}

double LoopTimer::avgTime() {
	return (double) totalTime() / (double) step;
}
